<?php

/*
 * PHP Utilities - Misc GS1 functions
 * Copyright 2023 - 2024 Daniel Marschall, ViaThinkSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Shows the GS1 Application Identifier for a given GS1 barcode
 * @param string $gs1 The GS1 barcode, e.g. 0012000000171
 * @return string The detected GS1 Application Identifier
 */
function gs1_barcode_show_appidentifier(string $gs1): string {
	// Lookup table taken from:  https://ref.gs1.org/ai/ (English setting)
	// Date: 01 Dec 2024 (522 entries)
	// The following code is used to re-generate the lookup table $aids ($raw is inserted via copy-paste from the browser)
	/*
	$raw = <<<EOF
	AI	Description	Format	Data title	FNC1 required?	Regular expression
	00	Serial Shipping Container Code (SSCC)	N2+N18	SSCC	No	^00(\d{18})$
	01	Global Trade Item Number (GTIN)	N2+N14	GTIN	No	^01(\d{14})$
	.......
	EOF;
	$out = "\tstatic \$aids = [ // [NoOidplusContextOk]\n";
	$count = 0;
	foreach (explode("\n",$raw) as $line) {
		$line = trim($line);
		if ($line == '') continue;
		if (strpos($line,"\t") === false) continue;
		list($prefix, $description) = explode("\t", $line);
		if ($prefix == "AI") continue;
		$description = str_replace('\\', '\\\\', $description);
		$description = str_replace("'", "\\'", $description);
		$out .= "\t\t'$prefix' => '$description',\n";
		$count++;
	}
	echo substr($out,0,strlen($out)-2) . "\n\t];\n// Count: $count\n";
	*/
	static $aids = [ // [NoOidplusContextOk]
		'00' => 'Serial Shipping Container Code (SSCC)',
		'01' => 'Global Trade Item Number (GTIN)',
		'02' => 'Identification of trade items contained in a logistic unit',
		'10' => 'Batch or lot number',
		'11' => 'Production date (YYMMDD)',
		'12' => 'Due date for amount on payment slip (YYMMDD)',
		'13' => 'Packaging date (YYMMDD)',
		'15' => 'Best before date (YYMMDD)',
		'16' => 'Sell by date (YYMMDD)',
		'17' => 'Expiration date (YYMMDD)',
		'20' => 'Internal product variant',
		'21' => 'Serial number',
		'22' => 'Consumer product variant',
		'235' => 'Third Party Controlled, Serialised Extension of Global Trade Item Number (GTIN) (TPX)',
		'240' => 'Additional product identification assigned by the manufacturer',
		'241' => 'Customer part number',
		'242' => 'Made-to-Order variation number',
		'243' => 'Packaging component number',
		'250' => 'Secondary serial number',
		'251' => 'Reference to source entity',
		'253' => 'Global Document Type Identifier (GDTI)',
		'254' => 'Global Location Number (GLN) extension component',
		'255' => 'Global Coupon Number (GCN)',
		'30' => 'Variable count of items (variable measure trade item)',
		'3100' => 'Net weight, kilograms (variable measure trade item)',
		'3101' => 'Net weight, kilograms (variable measure trade item)',
		'3102' => 'Net weight, kilograms (variable measure trade item)',
		'3103' => 'Net weight, kilograms (variable measure trade item)',
		'3104' => 'Net weight, kilograms (variable measure trade item)',
		'3105' => 'Net weight, kilograms (variable measure trade item)',
		'3110' => 'Length or first dimension, metres (variable measure trade item)',
		'3111' => 'Length or first dimension, metres (variable measure trade item)',
		'3112' => 'Length or first dimension, metres (variable measure trade item)',
		'3113' => 'Length or first dimension, metres (variable measure trade item)',
		'3114' => 'Length or first dimension, metres (variable measure trade item)',
		'3115' => 'Length or first dimension, metres (variable measure trade item)',
		'3120' => 'Width, diameter, or second dimension, metres (variable measure trade item)',
		'3121' => 'Width, diameter, or second dimension, metres (variable measure trade item)',
		'3122' => 'Width, diameter, or second dimension, metres (variable measure trade item)',
		'3123' => 'Width, diameter, or second dimension, metres (variable measure trade item)',
		'3124' => 'Width, diameter, or second dimension, metres (variable measure trade item)',
		'3125' => 'Width, diameter, or second dimension, metres (variable measure trade item)',
		'3130' => 'Depth, thickness, height, or third dimension, metres (variable measure trade item)',
		'3131' => 'Depth, thickness, height, or third dimension, metres (variable measure trade item)',
		'3132' => 'Depth, thickness, height, or third dimension, metres (variable measure trade item)',
		'3133' => 'Depth, thickness, height, or third dimension, metres (variable measure trade item)',
		'3134' => 'Depth, thickness, height, or third dimension, metres (variable measure trade item)',
		'3135' => 'Depth, thickness, height, or third dimension, metres (variable measure trade item)',
		'3140' => 'Area, square metres (variable measure trade item)',
		'3141' => 'Area, square metres (variable measure trade item)',
		'3142' => 'Area, square metres (variable measure trade item)',
		'3143' => 'Area, square metres (variable measure trade item)',
		'3144' => 'Area, square metres (variable measure trade item)',
		'3145' => 'Area, square metres (variable measure trade item)',
		'3150' => 'Net volume, litres (variable measure trade item)',
		'3151' => 'Net volume, litres (variable measure trade item)',
		'3152' => 'Net volume, litres (variable measure trade item)',
		'3153' => 'Net volume, litres (variable measure trade item)',
		'3154' => 'Net volume, litres (variable measure trade item)',
		'3155' => 'Net volume, litres (variable measure trade item)',
		'3160' => 'Net volume, cubic metres (variable measure trade item)',
		'3161' => 'Net volume, cubic metres (variable measure trade item)',
		'3162' => 'Net volume, cubic metres (variable measure trade item)',
		'3163' => 'Net volume, cubic metres (variable measure trade item)',
		'3164' => 'Net volume, cubic metres (variable measure trade item)',
		'3165' => 'Net volume, cubic metres (variable measure trade item)',
		'3200' => 'Net weight, pounds (variable measure trade item)',
		'3201' => 'Net weight, pounds (variable measure trade item)',
		'3202' => 'Net weight, pounds (variable measure trade item)',
		'3203' => 'Net weight, pounds (variable measure trade item)',
		'3204' => 'Net weight, pounds (variable measure trade item)',
		'3205' => 'Net weight, pounds (variable measure trade item)',
		'3210' => 'Length or first dimension, inches (variable measure trade item)',
		'3211' => 'Length or first dimension, inches (variable measure trade item)',
		'3212' => 'Length or first dimension, inches (variable measure trade item)',
		'3213' => 'Length or first dimension, inches (variable measure trade item)',
		'3214' => 'Length or first dimension, inches (variable measure trade item)',
		'3215' => 'Length or first dimension, inches (variable measure trade item)',
		'3220' => 'Length or first dimension, feet (variable measure trade item)',
		'3221' => 'Length or first dimension, feet (variable measure trade item)',
		'3222' => 'Length or first dimension, feet (variable measure trade item)',
		'3223' => 'Length or first dimension, feet (variable measure trade item)',
		'3224' => 'Length or first dimension, feet (variable measure trade item)',
		'3225' => 'Length or first dimension, feet (variable measure trade item)',
		'3230' => 'Length or first dimension, yards (variable measure trade item)',
		'3231' => 'Length or first dimension, yards (variable measure trade item)',
		'3232' => 'Length or first dimension, yards (variable measure trade item)',
		'3233' => 'Length or first dimension, yards (variable measure trade item)',
		'3234' => 'Length or first dimension, yards (variable measure trade item)',
		'3235' => 'Length or first dimension, yards (variable measure trade item)',
		'3240' => 'Width, diameter, or second dimension, inches (variable measure trade item)',
		'3241' => 'Width, diameter, or second dimension, inches (variable measure trade item)',
		'3242' => 'Width, diameter, or second dimension, inches (variable measure trade item)',
		'3243' => 'Width, diameter, or second dimension, inches (variable measure trade item)',
		'3244' => 'Width, diameter, or second dimension, inches (variable measure trade item)',
		'3245' => 'Width, diameter, or second dimension, inches (variable measure trade item)',
		'3250' => 'Width, diameter, or second dimension, feet (variable measure trade item)',
		'3251' => 'Width, diameter, or second dimension, feet (variable measure trade item)',
		'3252' => 'Width, diameter, or second dimension, feet (variable measure trade item)',
		'3253' => 'Width, diameter, or second dimension, feet (variable measure trade item)',
		'3254' => 'Width, diameter, or second dimension, feet (variable measure trade item)',
		'3255' => 'Width, diameter, or second dimension, feet (variable measure trade item)',
		'3260' => 'Width, diameter, or second dimension, yards (variable measure trade item)',
		'3261' => 'Width, diameter, or second dimension, yards (variable measure trade item)',
		'3262' => 'Width, diameter, or second dimension, yards (variable measure trade item)',
		'3263' => 'Width, diameter, or second dimension, yards (variable measure trade item)',
		'3264' => 'Width, diameter, or second dimension, yards (variable measure trade item)',
		'3265' => 'Width, diameter, or second dimension, yards (variable measure trade item)',
		'3270' => 'Depth, thickness, height, or third dimension, inches (variable measure trade item)',
		'3271' => 'Depth, thickness, height, or third dimension, inches (variable measure trade item)',
		'3272' => 'Depth, thickness, height, or third dimension, inches (variable measure trade item)',
		'3273' => 'Depth, thickness, height, or third dimension, inches (variable measure trade item)',
		'3274' => 'Depth, thickness, height, or third dimension, inches (variable measure trade item)',
		'3275' => 'Depth, thickness, height, or third dimension, inches (variable measure trade item)',
		'3280' => 'Depth, thickness, height, or third dimension, feet (variable measure trade item)',
		'3281' => 'Depth, thickness, height, or third dimension, feet (variable measure trade item)',
		'3282' => 'Depth, thickness, height, or third dimension, feet (variable measure trade item)',
		'3283' => 'Depth, thickness, height, or third dimension, feet (variable measure trade item)',
		'3284' => 'Depth, thickness, height, or third dimension, feet (variable measure trade item)',
		'3285' => 'Depth, thickness, height, or third dimension, feet (variable measure trade item)',
		'3290' => 'Depth, thickness, height, or third dimension, yards (variable measure trade item)',
		'3291' => 'Depth, thickness, height, or third dimension, yards (variable measure trade item)',
		'3292' => 'Depth, thickness, height, or third dimension, yards (variable measure trade item)',
		'3293' => 'Depth, thickness, height, or third dimension, yards (variable measure trade item)',
		'3294' => 'Depth, thickness, height, or third dimension, yards (variable measure trade item)',
		'3295' => 'Depth, thickness, height, or third dimension, yards (variable measure trade item)',
		'3300' => 'Logistic weight, kilograms',
		'3301' => 'Logistic weight, kilograms',
		'3302' => 'Logistic weight, kilograms',
		'3303' => 'Logistic weight, kilograms',
		'3304' => 'Logistic weight, kilograms',
		'3305' => 'Logistic weight, kilograms',
		'3310' => 'Length or first dimension, metres',
		'3311' => 'Length or first dimension, metres',
		'3312' => 'Length or first dimension, metres',
		'3313' => 'Length or first dimension, metres',
		'3314' => 'Length or first dimension, metres',
		'3315' => 'Length or first dimension, metres',
		'3320' => 'Width, diameter, or second dimension, metres',
		'3321' => 'Width, diameter, or second dimension, metres',
		'3322' => 'Width, diameter, or second dimension, metres',
		'3323' => 'Width, diameter, or second dimension, metres',
		'3324' => 'Width, diameter, or second dimension, metres',
		'3325' => 'Width, diameter, or second dimension, metres',
		'3330' => 'Depth, thickness, height, or third dimension, metres',
		'3331' => 'Depth, thickness, height, or third dimension, metres',
		'3332' => 'Depth, thickness, height, or third dimension, metres',
		'3333' => 'Depth, thickness, height, or third dimension, metres',
		'3334' => 'Depth, thickness, height, or third dimension, metres',
		'3335' => 'Depth, thickness, height, or third dimension, metres',
		'3340' => 'Area, square metres',
		'3341' => 'Area, square metres',
		'3342' => 'Area, square metres',
		'3343' => 'Area, square metres',
		'3344' => 'Area, square metres',
		'3345' => 'Area, square metres',
		'3350' => 'Logistic volume, litres',
		'3351' => 'Logistic volume, litres',
		'3352' => 'Logistic volume, litres',
		'3353' => 'Logistic volume, litres',
		'3354' => 'Logistic volume, litres',
		'3355' => 'Logistic volume, litres',
		'3360' => 'Logistic volume, cubic metres',
		'3361' => 'Logistic volume, cubic metres',
		'3362' => 'Logistic volume, cubic metres',
		'3363' => 'Logistic volume, cubic metres',
		'3364' => 'Logistic volume, cubic metres',
		'3365' => 'Logistic volume, cubic metres',
		'3370' => 'Kilograms per square metre',
		'3371' => 'Kilograms per square metre',
		'3372' => 'Kilograms per square metre',
		'3373' => 'Kilograms per square metre',
		'3374' => 'Kilograms per square metre',
		'3375' => 'Kilograms per square metre',
		'3400' => 'Logistic weight, pounds',
		'3401' => 'Logistic weight, pounds',
		'3402' => 'Logistic weight, pounds',
		'3403' => 'Logistic weight, pounds',
		'3404' => 'Logistic weight, pounds',
		'3405' => 'Logistic weight, pounds',
		'3410' => 'Length or first dimension, inches',
		'3411' => 'Length or first dimension, inches',
		'3412' => 'Length or first dimension, inches',
		'3413' => 'Length or first dimension, inches',
		'3414' => 'Length or first dimension, inches',
		'3415' => 'Length or first dimension, inches',
		'3420' => 'Length or first dimension, feet',
		'3421' => 'Length or first dimension, feet',
		'3422' => 'Length or first dimension, feet',
		'3423' => 'Length or first dimension, feet',
		'3424' => 'Length or first dimension, feet',
		'3425' => 'Length or first dimension, feet',
		'3430' => 'Length or first dimension, yards',
		'3431' => 'Length or first dimension, yards',
		'3432' => 'Length or first dimension, yards',
		'3433' => 'Length or first dimension, yards',
		'3434' => 'Length or first dimension, yards',
		'3435' => 'Length or first dimension, yards',
		'3440' => 'Width, diameter, or second dimension, inches',
		'3441' => 'Width, diameter, or second dimension, inches',
		'3442' => 'Width, diameter, or second dimension, inches',
		'3443' => 'Width, diameter, or second dimension, inches',
		'3444' => 'Width, diameter, or second dimension, inches',
		'3445' => 'Width, diameter, or second dimension, inches',
		'3450' => 'Width, diameter, or second dimension, feet',
		'3451' => 'Width, diameter, or second dimension, feet',
		'3452' => 'Width, diameter, or second dimension, feet',
		'3453' => 'Width, diameter, or second dimension, feet',
		'3454' => 'Width, diameter, or second dimension, feet',
		'3455' => 'Width, diameter, or second dimension, feet',
		'3460' => 'Width, diameter, or second dimension, yard',
		'3461' => 'Width, diameter, or second dimension, yard',
		'3462' => 'Width, diameter, or second dimension, yard',
		'3463' => 'Width, diameter, or second dimension, yard',
		'3464' => 'Width, diameter, or second dimension, yard',
		'3465' => 'Width, diameter, or second dimension, yard',
		'3470' => 'Depth, thickness, height, or third dimension, inches',
		'3471' => 'Depth, thickness, height, or third dimension, inches',
		'3472' => 'Depth, thickness, height, or third dimension, inches',
		'3473' => 'Depth, thickness, height, or third dimension, inches',
		'3474' => 'Depth, thickness, height, or third dimension, inches',
		'3475' => 'Depth, thickness, height, or third dimension, inches',
		'3480' => 'Depth, thickness, height, or third dimension, feet',
		'3481' => 'Depth, thickness, height, or third dimension, feet',
		'3482' => 'Depth, thickness, height, or third dimension, feet',
		'3483' => 'Depth, thickness, height, or third dimension, feet',
		'3484' => 'Depth, thickness, height, or third dimension, feet',
		'3485' => 'Depth, thickness, height, or third dimension, feet',
		'3490' => 'Depth, thickness, height, or third dimension, yards',
		'3491' => 'Depth, thickness, height, or third dimension, yards',
		'3492' => 'Depth, thickness, height, or third dimension, yards',
		'3493' => 'Depth, thickness, height, or third dimension, yards',
		'3494' => 'Depth, thickness, height, or third dimension, yards',
		'3495' => 'Depth, thickness, height, or third dimension, yards',
		'3500' => 'Area, square inches',
		'3501' => 'Area, square inches',
		'3502' => 'Area, square inches',
		'3503' => 'Area, square inches',
		'3504' => 'Area, square inches',
		'3505' => 'Area, square inches',
		'3510' => 'Area, square feet (variable measure trade item)',
		'3511' => 'Area, square feet (variable measure trade item)',
		'3512' => 'Area, square feet (variable measure trade item)',
		'3513' => 'Area, square feet (variable measure trade item)',
		'3514' => 'Area, square feet (variable measure trade item)',
		'3515' => 'Area, square feet (variable measure trade item)',
		'3520' => 'Area, square yards (variable measure trade item)',
		'3521' => 'Area, square yards (variable measure trade item)',
		'3522' => 'Area, square yards (variable measure trade item)',
		'3523' => 'Area, square yards (variable measure trade item)',
		'3524' => 'Area, square yards (variable measure trade item)',
		'3525' => 'Area, square yards (variable measure trade item)',
		'3530' => 'Area, square inches',
		'3531' => 'Area, square inches',
		'3532' => 'Area, square inches',
		'3533' => 'Area, square inches',
		'3534' => 'Area, square inches',
		'3535' => 'Area, square inches',
		'3540' => 'Area, square feet',
		'3541' => 'Area, square feet',
		'3542' => 'Area, square feet',
		'3543' => 'Area, square feet',
		'3544' => 'Area, square feet',
		'3545' => 'Area, square feet',
		'3550' => 'Area, square yards',
		'3551' => 'Area, square yards',
		'3552' => 'Area, square yards',
		'3553' => 'Area, square yards',
		'3554' => 'Area, square yards',
		'3555' => 'Area, square yards',
		'3560' => 'Net weight, troy ounces (variable measure trade item)',
		'3561' => 'Net weight, troy ounces (variable measure trade item)',
		'3562' => 'Net weight, troy ounces (variable measure trade item)',
		'3563' => 'Net weight, troy ounces (variable measure trade item)',
		'3564' => 'Net weight, troy ounces (variable measure trade item)',
		'3565' => 'Net weight, troy ounces (variable measure trade item)',
		'3570' => 'Net weight (or volume), ounces (variable measure trade item)',
		'3571' => 'Net weight (or volume), ounces (variable measure trade item)',
		'3572' => 'Net weight (or volume), ounces (variable measure trade item)',
		'3573' => 'Net weight (or volume), ounces (variable measure trade item)',
		'3574' => 'Net weight (or volume), ounces (variable measure trade item)',
		'3575' => 'Net weight (or volume), ounces (variable measure trade item)',
		'3600' => 'Net volume, quarts (variable measure trade item)',
		'3601' => 'Net volume, quarts (variable measure trade item)',
		'3602' => 'Net volume, quarts (variable measure trade item)',
		'3603' => 'Net volume, quarts (variable measure trade item)',
		'3604' => 'Net volume, quarts (variable measure trade item)',
		'3605' => 'Net volume, quarts (variable measure trade item)',
		'3610' => 'Net volume, gallons U.S. (variable measure trade item)',
		'3611' => 'Net volume, gallons U.S. (variable measure trade item)',
		'3612' => 'Net volume, gallons U.S. (variable measure trade item)',
		'3613' => 'Net volume, gallons U.S. (variable measure trade item)',
		'3614' => 'Net volume, gallons U.S. (variable measure trade item)',
		'3615' => 'Net volume, gallons U.S. (variable measure trade item)',
		'3620' => 'Logistic volume, quarts',
		'3621' => 'Logistic volume, quarts',
		'3622' => 'Logistic volume, quarts',
		'3623' => 'Logistic volume, quarts',
		'3624' => 'Logistic volume, quarts',
		'3625' => 'Logistic volume, quarts',
		'3630' => 'Logistic volume, gallons U.S.',
		'3631' => 'Logistic volume, gallons U.S.',
		'3632' => 'Logistic volume, gallons U.S.',
		'3633' => 'Logistic volume, gallons U.S.',
		'3634' => 'Logistic volume, gallons U.S.',
		'3635' => 'Logistic volume, gallons U.S.',
		'3640' => 'Net volume, cubic inches (variable measure trade item)',
		'3641' => 'Net volume, cubic inches (variable measure trade item)',
		'3642' => 'Net volume, cubic inches (variable measure trade item)',
		'3643' => 'Net volume, cubic inches (variable measure trade item)',
		'3644' => 'Net volume, cubic inches (variable measure trade item)',
		'3645' => 'Net volume, cubic inches (variable measure trade item)',
		'3650' => 'Net volume, cubic feet (variable measure trade item)',
		'3651' => 'Net volume, cubic feet (variable measure trade item)',
		'3652' => 'Net volume, cubic feet (variable measure trade item)',
		'3653' => 'Net volume, cubic feet (variable measure trade item)',
		'3654' => 'Net volume, cubic feet (variable measure trade item)',
		'3655' => 'Net volume, cubic feet (variable measure trade item)',
		'3660' => 'Net volume, cubic yards (variable measure trade item)',
		'3661' => 'Net volume, cubic yards (variable measure trade item)',
		'3662' => 'Net volume, cubic yards (variable measure trade item)',
		'3663' => 'Net volume, cubic yards (variable measure trade item)',
		'3664' => 'Net volume, cubic yards (variable measure trade item)',
		'3665' => 'Net volume, cubic yards (variable measure trade item)',
		'3670' => 'Logistic volume, cubic inches',
		'3671' => 'Logistic volume, cubic inches',
		'3672' => 'Logistic volume, cubic inches',
		'3673' => 'Logistic volume, cubic inches',
		'3674' => 'Logistic volume, cubic inches',
		'3675' => 'Logistic volume, cubic inches',
		'3680' => 'Logistic volume, cubic feet',
		'3681' => 'Logistic volume, cubic feet',
		'3682' => 'Logistic volume, cubic feet',
		'3683' => 'Logistic volume, cubic feet',
		'3684' => 'Logistic volume, cubic feet',
		'3685' => 'Logistic volume, cubic feet',
		'3690' => 'Logistic volume, cubic yards',
		'3691' => 'Logistic volume, cubic yards',
		'3692' => 'Logistic volume, cubic yards',
		'3693' => 'Logistic volume, cubic yards',
		'3694' => 'Logistic volume, cubic yards',
		'3695' => 'Logistic volume, cubic yards',
		'37' => 'Count of trade items or trade item pieces contained in a logistic unit',
		'3900' => 'Amount payable or coupon value - Single monetary area',
		'3901' => 'Amount payable or coupon value - Single monetary area',
		'3902' => 'Amount payable or coupon value - Single monetary area',
		'3903' => 'Amount payable or coupon value - Single monetary area',
		'3904' => 'Amount payable or coupon value - Single monetary area',
		'3905' => 'Amount payable or coupon value - Single monetary area',
		'3906' => 'Amount payable or coupon value - Single monetary area',
		'3907' => 'Amount payable or coupon value - Single monetary area',
		'3908' => 'Amount payable or coupon value - Single monetary area',
		'3909' => 'Amount payable or coupon value - Single monetary area',
		'3910' => 'Amount payable and ISO currency code',
		'3911' => 'Amount payable and ISO currency code',
		'3912' => 'Amount payable and ISO currency code',
		'3913' => 'Amount payable and ISO currency code',
		'3914' => 'Amount payable and ISO currency code',
		'3915' => 'Amount payable and ISO currency code',
		'3916' => 'Amount payable and ISO currency code',
		'3917' => 'Amount payable and ISO currency code',
		'3918' => 'Amount payable and ISO currency code',
		'3919' => 'Amount payable and ISO currency code',
		'3920' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3921' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3922' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3923' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3924' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3925' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3926' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3927' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3928' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3929' => 'Amount payable for a variable measure trade item – Single monetary area',
		'3930' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3931' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3932' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3933' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3934' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3935' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3936' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3937' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3938' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3939' => 'Amount payable for a variable measure trade item and ISO currency code',
		'3940' => 'Percentage discount of a coupon',
		'3941' => 'Percentage discount of a coupon',
		'3942' => 'Percentage discount of a coupon',
		'3943' => 'Percentage discount of a coupon',
		'3950' => 'Amount Payable per unit of measure single monetary area (variable measure trade item)',
		'3951' => 'Amount Payable per unit of measure single monetary area (variable measure trade item)',
		'3952' => 'Amount Payable per unit of measure single monetary area (variable measure trade item)',
		'3953' => 'Amount Payable per unit of measure single monetary area (variable measure trade item)',
		'3954' => 'Amount Payable per unit of measure single monetary area (variable measure trade item)',
		'3955' => 'Amount Payable per unit of measure single monetary area (variable measure trade item)',
		'400' => 'Customer’s purchase order number',
		'401' => 'Global Identification Number for Consignment (GINC)',
		'402' => 'Global Shipment Identification Number (GSIN)',
		'403' => 'Routing code',
		'410' => 'Ship to / Deliver to Global Location Number (GLN)',
		'411' => 'Bill to / Invoice to Global Location Number (GLN)',
		'412' => 'Purchased from Global Location Number (GLN)',
		'413' => 'Ship for / Deliver for - Forward to Global Location Number (GLN)',
		'414' => 'Identification of a physical location - Global Location Number (GLN)',
		'415' => 'Global Location Number (GLN) of the invoicing party',
		'416' => 'Global Location Number (GLN) of the production or service location',
		'417' => 'Party Global Location Number (GLN)',
		'420' => 'Ship to / Deliver to postal code within a single postal authority',
		'421' => 'Ship to / Deliver to postal code with ISO country code',
		'422' => 'Country of origin of a trade item',
		'423' => 'Country of initial processing',
		'424' => 'Country of processing',
		'425' => 'Country of disassembly',
		'426' => 'Country covering full process chain',
		'427' => 'Country subdivision of origin code for a trade item',
		'4300' => 'Ship-to / Deliver-to Company name',
		'4301' => 'Ship-to / Deliver-to contact name',
		'4302' => 'Ship-to / Deliver-to address line 1',
		'4303' => 'Ship-to / Deliver-to address line 2',
		'4304' => 'Ship-to / Deliver-to suburb',
		'4305' => 'Ship-to / Deliver-to locality',
		'4306' => 'Ship-to / Deliver-to region',
		'4307' => 'Ship-to / Deliver-to country code',
		'4308' => 'Ship-to / Deliver-to telephone number',
		'4309' => 'Ship-to / Deliver-to GEO location',
		'4310' => 'Return-to company name',
		'4311' => 'Return-to contact',
		'4312' => 'Return-to address line 1',
		'4313' => 'Return-to address line 2',
		'4314' => 'Return-to suburb',
		'4315' => 'Return-to locality',
		'4316' => 'Return-to region',
		'4317' => 'Return-to country code',
		'4318' => 'Return-to postal code',
		'4319' => 'Return-to telephone number',
		'4320' => 'Service code description',
		'4321' => 'Dangerous goods flag',
		'4322' => 'Authority to leave',
		'4323' => 'Signature required flag',
		'4324' => 'Not before delivery date time',
		'4325' => 'Not after delivery date time',
		'4326' => 'Release date',
		'4330' => 'Maximum temperature in Fahrenheit',
		'4331' => 'Maximum temperature in Celsius',
		'4332' => 'Minimum temperature in Fahrenheit',
		'4333' => 'Minimum temperature in Celsius',
		'7001' => 'NATO Stock Number (NSN)',
		'7002' => 'UNECE meat carcasses and cuts classification',
		'7003' => 'Expiration date and time',
		'7004' => 'Active potency',
		'7005' => 'Catch area',
		'7006' => 'First freeze date',
		'7007' => 'Harvest date',
		'7008' => 'Species for fishery purposes',
		'7009' => 'Fishing gear type',
		'7010' => 'Production method',
		'7011' => 'Test by date',
		'7020' => 'Refurbishment lot ID',
		'7021' => 'Functional status',
		'7022' => 'Revision status',
		'7023' => 'Global Individual Asset Identifier (GIAI) of an assembly',
		'7030' => 'Number of processor with three-digit ISO country code',
		'7031' => 'Number of processor with three-digit ISO country code',
		'7032' => 'Number of processor with three-digit ISO country code',
		'7033' => 'Number of processor with three-digit ISO country code',
		'7034' => 'Number of processor with three-digit ISO country code',
		'7035' => 'Number of processor with three-digit ISO country code',
		'7036' => 'Number of processor with three-digit ISO country code',
		'7037' => 'Number of processor with three-digit ISO country code',
		'7038' => 'Number of processor with three-digit ISO country code',
		'7039' => 'Number of processor with three-digit ISO country code',
		'7040' => 'GS1 UIC with Extension 1 and Importer index',
		'710' => 'National Healthcare Reimbursement Number (NHRN) - Germany PZN',
		'711' => 'National Healthcare Reimbursement Number (NHRN) - France CIP',
		'712' => 'National Healthcare Reimbursement Number (NHRN) - Spain CN',
		'713' => 'National Healthcare Reimbursement Number (NHRN) - Brasil DRN',
		'714' => 'National Healthcare Reimbursement Number (NHRN) - Portugal AIM',
		'715' => 'National Healthcare Reimbursement Number (NHRN) - United States of America NDC',
		'7230' => 'Certification reference',
		'7231' => 'Certification reference',
		'7232' => 'Certification reference',
		'7233' => 'Certification reference',
		'7234' => 'Certification reference',
		'7235' => 'Certification reference',
		'7236' => 'Certification reference',
		'7237' => 'Certification reference',
		'7238' => 'Certification reference',
		'7239' => 'Certification reference',
		'7240' => 'Protocol ID',
		'7241' => 'AIDC media type',
		'7242' => 'Version Control Number (VCN)',
		'8001' => 'Roll products (width, length, core diameter, direction, splices)',
		'8002' => 'Cellular mobile telephone identifier',
		'8003' => 'Global Returnable Asset Identifier (GRAI)',
		'8004' => 'Global Individual Asset Identifier (GIAI)',
		'8005' => 'Price per unit of measure',
		'8006' => 'Identification of an individual trade item (ITIP) piece',
		'8007' => 'International Bank Account Number (IBAN)',
		'8008' => 'Date and time of production',
		'8009' => 'Optically Readable Sensor Indicator',
		'8010' => 'Component/Part Identifier (CPID)',
		'8011' => 'Component/Part Identifier serial number (CPID SERIAL)',
		'8012' => 'Software version',
		'8013' => 'Global Model Number (GMN)',
		'8017' => 'Global Service Relation Number (GSRN) to identify the relationship between an organisation offering services and the provider of services',
		'8018' => 'Global Service Relation Number (GSRN) to identify the relationship between an organisation offering services and the recipient of services',
		'8019' => 'Service Relation Instance Number (SRIN)',
		'8020' => 'Payment slip reference number',
		'8026' => 'Identification of pieces of a trade item (ITIP) contained in a logistic unit',
		'8030' => 'Digital Signature (DigSig)',
		'8110' => 'Coupon code identification for use in North America',
		'8111' => 'Loyalty points of a coupon',
		'8112' => 'Positive offer file coupon code identification for use in North America',
		'8200' => 'Extended Packaging URL',
		'90' => 'Information mutually agreed between trading partners',
		'91' => 'Company internal information',
		'92' => 'Company internal information',
		'93' => 'Company internal information',
		'94' => 'Company internal information',
		'95' => 'Company internal information',
		'96' => 'Company internal information',
		'97' => 'Company internal information',
		'98' => 'Company internal information',
		'99' => 'Company internal information'
	];

	foreach ($aids as $aid => $description) {
		if (strpos($gs1, $aid) === 0) {
			return "$aid = $description";
		}
	}
	return "Unknown";
}
